/*
 * kgraft_patch_bsc960329
 *
 * Fix for CVE-2015-8660, bsc#960329
 *
 *  Upstream commit:
 *  acff81ec2c79492b180fade3c2894425cd35a545
 *
 *  SLE12-SP1 commits:
 *  97002d5956250cf79c5a641e8161e73adf91ac42
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/fs.h>

static int (*kgr_ovl_want_write)(struct dentry *dentry);
static int (*kgr_ovl_copy_up)(struct dentry *dentry);
static struct dentry *(*kgr_ovl_dentry_upper)(struct dentry *dentry);
static void (*kgr_ovl_drop_write)(struct dentry *dentry);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "ovl_want_write", (void *)&kgr_ovl_want_write },
	{ "ovl_copy_up", (void *)&kgr_ovl_copy_up },
	{ "ovl_dentry_upper", (void *)&kgr_ovl_dentry_upper },
	{ "ovl_drop_write", (void *)&kgr_ovl_drop_write },
};

/* Patched function */
int kgr_ovl_setattr(struct dentry *dentry, struct iattr *attr)
{
	int err;
	struct dentry *upperdentry;

	err = kgr_ovl_want_write(dentry);
	if (err)
		goto out;

	/*
	 * Fix CVE-2015-8660
	 *  -upperdentry = kgr_ovl_dentry_upper(dentry);
	 *  -if (upperdentry) {
	 *  +next 4 rows
	 */
	err = kgr_ovl_copy_up(dentry);
	if (!err) {
		upperdentry = kgr_ovl_dentry_upper(dentry);

		mutex_lock(&upperdentry->d_inode->i_mutex);
		err = notify_change(upperdentry, attr, NULL);
		mutex_unlock(&upperdentry->d_inode->i_mutex);
	/*
	 * Fix CVE-2015-8660
	 *  -} else {
	 *  -err = ovl_copy_up_last(dentry, attr, false);
	 */
	}
	kgr_ovl_drop_write(dentry);
out:
	return err;
}

static int kgr_patch_bsc960329_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc960329_module_notify(struct notifier_block *nb,
                unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, "overlay"))
		return 0;

	ret = kgr_patch_bsc960329_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc960329_module_nb = {
	.notifier_call = kgr_patch_bsc960329_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc960329_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module("overlay")) {
		ret = kgr_patch_bsc960329_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc960329_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc960329_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc960329_module_nb);
}
